"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod
from typing import Any, Callable

class IResolver(ABC):
    @abstractmethod
    def resolve_anonymous[T](self, constructor: Callable[..., T], overrides: dict[str, Any], binding: Any | None = ...) -> T:
        ...
    
    @abstractmethod
    def resolve[T](self, interface: type[T], overrides: dict[str, Any]) -> T:
        ...
    





"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod
from typing import Any, Callable, Self

class IServiceScope(ABC):
    @abstractmethod
    def get[T](self, interface: type[T], **overrides: Any) -> T:
        ...
    
    @abstractmethod
    def get_instance[T](self, interface: type[T]) -> T | None:
        ...
    
    @abstractmethod
    def add_instance[T](self, interface: type[T], instance: T) -> None:
        ...
    
    @abstractmethod
    def __getitem__[T](self, key: type[T]) -> Callable[..., T]:
        ...
    
    @abstractmethod
    def __enter__(self) -> Self:
        ...
    
    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...
    





from abc import ABC, abstractmethod
from typing import Any, Self


class IServiceCache(ABC):
    @abstractmethod
    def get[T](self, interface: type[T]) -> T | None: ...

    @abstractmethod
    def add[T](self, interface: type[T], instance: T) -> None: ...


"""
This type stub file was generated by pyright.
"""

from typing import Any, override
from pyjudo.core import IServiceCache

class ServiceCache(IServiceCache):
    def __init__(self, initial: dict[type[Any], Any] | None = ...) -> None:
        ...
    
    @override
    def get[T](self, interface: type[T]) -> T | None:
        ...
    
    @override
    def add[T](self, interface: type[T], instance: T) -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod
from .iservice_scope import IServiceScope

class IScopeStack(ABC):
    @abstractmethod
    def push(self, scope: IServiceScope) -> None:
        ...
    
    @abstractmethod
    def pop(self) -> None:
        ...
    
    @abstractmethod
    def get_current(self) -> IServiceScope | None:
        ...
    





from abc import ABC, abstractmethod
from typing import Any, Callable, Self


class IServiceScope(ABC):
    @abstractmethod
    def get[T](self, interface: type[T], **overrides: Any) -> T: ...

    @abstractmethod
    def get_instance[T](self, interface: type[T]) -> T | None: ...

    @abstractmethod
    def add_instance[T](self, interface: type[T], instance: T) -> None: ...

    @abstractmethod
    def __getitem__[T](self, key: type[T]) -> Callable[..., T]: ...

    @abstractmethod
    def __enter__(self) -> Self: ...

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None: ...


from abc import ABC
import unittest
from unittest.mock import MagicMock, Mock
from typing import Any, Callable

from pyjudo.core import IScopeStack, IServiceCache, IServiceEntryCollection, IServiceScope, IResolver, ServiceLife
from pyjudo.exceptions import ServiceRegistrationError, ServiceResolutionError, ServiceTypeError
from pyjudo.factory import FactoryProxy
from pyjudo.service_container import ServiceContainer


class IMockService(ABC): ...

class MockService(IMockService): ...


class Test_ServiceContainer(unittest.TestCase):
    def setUp(self):
        self.mock_service_entry_collection: IServiceEntryCollection = MagicMock(spec=IServiceEntryCollection)
        self.mock_singleton_cache: IServiceCache = Mock(spec=IServiceCache)
        self.mock_scope_stack: IScopeStack = Mock(spec=IScopeStack)
        self.mock_resolver: IResolver = Mock(spec=IResolver)
        self.mock_scope_factory: Callable[..., IServiceScope] = Mock(return_value=Mock(spec=IServiceScope))

        self.container = ServiceContainer(
            service_entry_collection=self.mock_service_entry_collection,
            singleton_cache=self.mock_singleton_cache,
            scope_stack=self.mock_scope_stack,
            scope_factory=self.mock_scope_factory,
            resolver=self.mock_resolver,
        )

    def test_register_valid_class(self):
        interface = IMockService
        constructor = MockService
        self.container.register(interface, constructor, ServiceLife.SINGLETON)

        # Called twice, the first time during construction when the service container adds itself.
        self.assertEqual(self.mock_service_entry_collection.set.call_count, 2)
        entry = self.mock_service_entry_collection.set.call_args[0][1]
        self.assertEqual(entry.constructor, constructor)
        self.assertEqual(entry.lifetime, ServiceLife.SINGLETON)

    def test_register_invalid_class(self):
        interface = IMockService
        constructor = Mock()

        with self.assertRaises(ServiceRegistrationError):
            self.container.register(interface, constructor, ServiceLife.SINGLETON)

    def test_register_callable_with_annotation(self):
        def constructor() -> MockService:
            return MockService()

        self.container.register(IMockService, constructor, ServiceLife.SINGLETON)

        # Called twice, the first time during construction when the service container adds itself.
        self.assertEqual(self.mock_service_entry_collection.set.call_count, 2)

    def test_register_callable_missing_annotation(self):
        def constructor():
            return MockService()

        with self.assertRaises(ServiceRegistrationError):
            self.container.register(IMockService, constructor, ServiceLife.SINGLETON)

    def test_inject_decorator(self):
        @self.container.inject
        def func(arg1: Any, arg2: Any) -> Any:
            pass

        func(arg1=1, arg2=2)
        self.mock_resolver.resolve_anonymous.assert_called_once()

    def test_unregister(self):
        interface = IMockService
        constructor = MockService
        self.container.unregister(interface)
        self.mock_service_entry_collection.remove.assert_called_once_with(interface)

    def test_add_transient(self):
        interface = IMockService
        constructor = MockService
        self.container.add_transient(interface, constructor)
        # Called twice, the first time during construction when the service container adds itself.
        self.assertEqual(self.mock_service_entry_collection.set.call_count, 2)

    def test_add_scoped(self):
        interface = IMockService
        constructor = MockService
        self.container.add_scoped(interface, constructor)
        # Called twice, the first time during construction when the service container adds itself.
        self.assertEqual(self.mock_service_entry_collection.set.call_count, 2)

    def test_add_singleton(self):
        interface = IMockService
        constructor = MockService
        self.container.add_singleton(interface, constructor)
        # Called twice, the first time during construction when the service container adds itself.
        self.assertEqual(self.mock_service_entry_collection.set.call_count, 2)

    def test_get_valid_service(self):
        interface = IMockService
        service_instance = MockService()
        self.mock_resolver.resolve.return_value = service_instance

        result = self.container.get(interface)
        self.assertIs(result, service_instance)

    def test_get_invalid_service_type(self):
        interface = IMockService
        service_instance = Mock()
        self.mock_resolver.resolve.return_value = service_instance

        with self.assertRaises(ServiceTypeError):
            self.container.get(interface)

    def test_get_factory_registered_service(self):
        interface = IMockService
        self.mock_service_entry_collection.__contains__.return_value = True
        factory = self.container.get_factory(interface)
        self.assertIsInstance(factory, FactoryProxy)

    def test_get_factory_unregistered_service(self):
        interface = IMockService
        self.mock_service_entry_collection.__contains__.return_value = False

        with self.assertRaises(ServiceResolutionError):
            self.container.get_factory(interface)

    def test_is_registered(self):
        interface = IMockService
        self.mock_service_entry_collection.__contains__.return_value = True
        self.assertTrue(self.container.is_registered(interface))

    def test_create_scope(self):
        scope = self.container.create_scope()
        self.mock_scope_factory.assert_called_once_with(self.mock_scope_stack, self.mock_resolver)
        self.assertIsInstance(scope, IServiceScope)

if __name__ == "__main__":
    unittest.main()



"""
PyJudo: Simple dependency injection in Python.
"""

from pyjudo.core import IServiceContainer, ServiceLife
from .factory import Factory
from .service_container import ServiceContainer as ServiceContainerBase

def ServiceContainer() -> ServiceContainerBase:
    ...

__all__ = ["Factory", "ServiceContainer", "IServiceContainer", "ServiceLife"]



"""
This type stub file was generated by pyright.
"""

from typing import Any, Callable, override
from pyjudo.core import IResolver, IScopeStack, IServiceCache, IServiceScope

class ServiceScope(IServiceScope):
    """
    Represents a scope for services.
    """
    def __init__(self, cache: IServiceCache, stack: IScopeStack, resolver: IResolver) -> None:
        ...
    
    @override
    def get[T](self, interface: type[T], **overrides: Any) -> T:
        ...
    
    @override
    def get_instance[T](self, interface: type[T]) -> T | None:
        ...
    
    @override
    def add_instance[T](self, interface: type[T], instance: T) -> None:
        ...
    
    @override
    def __getitem__[T](self, key: type[T]) -> Callable[..., T]:
        ...
    
    @override
    def __enter__(self): # -> Self:
        ...
    
    @override
    def __exit__(self, _exc_type, _exc_value, _traceback): # -> None:
        ...
    





import threading
from typing import Any
import unittest
from unittest.mock import Mock

from pyjudo.service_cache import ServiceCache


class Test_ServiceCache(unittest.TestCase):
    def setUp(self):
        self.cache: ServiceCache = ServiceCache() # pyright: ignore[reportUninitializedInstanceVariable]

    def test_get_returns_none_when_not_in_cache(self):
        result = self.cache.get(Mock)
        self.assertIsNone(result)

    def test_add_and_get_instance(self):
        mock_interface = Mock()
        mock_instance = Mock()

        self.cache.add(mock_interface, mock_instance)
        result = self.cache.get(mock_interface)

        self.assertIs(result, mock_instance)

    def test_override_existing_instance(self):
        mock_interface = Mock()
        old_instance = Mock()
        new_instance = Mock()

        self.cache.add(mock_interface, old_instance)
        self.cache.add(mock_interface, new_instance)

        result = self.cache.get(mock_interface)
        self.assertIs(result, new_instance)

    def test_initial_cache(self):
        initial_data: dict[type, Any] = {str: "string_instance", int: 42}
        cache = ServiceCache(initial=initial_data)

        self.assertEqual(cache.get(str), "string_instance")
        self.assertEqual(cache.get(int), 42)
        self.assertIsNone(cache.get(float))

    def test_thread_safety(self):
        def worker(cache: ServiceCache, interface: type, instance: Any):
            cache.add(interface, instance)
            result = cache.get(interface)
            self.assertIs(result, instance)

        threads = []
        for i in range(10):
            thread = threading.Thread(target=worker, args=(self.cache, type(f"MockInterface{i}", (), {}), f"Instance{i}"))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        # Verify cache consistency
        self.assertEqual(len(self.cache._cache), 10)  # Accessing internal cache for validation

    


from abc import ABC, abstractmethod
from typing import Any, Callable


class IResolver(ABC):
    @abstractmethod
    def resolve_anonymous[T](self, constructor: Callable[..., T], overrides: dict[str, Any], binding: Any | None = None) -> T: ...

    @abstractmethod
    def resolve[T](self, interface: type[T], overrides: dict[str, Any]) -> T: ...


import unittest
from unittest.mock import Mock
import threading

from pyjudo.core import ServiceEntry
from pyjudo.exceptions import ServiceRegistrationError, ServiceResolutionError
from pyjudo.service_entry_collection import ServiceEntryCollection


class Test_ServiceEntryCollection(unittest.TestCase):
    def setUp(self):
        self.collection = ServiceEntryCollection()

    def test_set_and_get_service(self):
        interface = Mock()
        entry = ServiceEntry(constructor=Mock(), lifetime=Mock())

        self.collection.set(interface, entry)
        retrieved_entry = self.collection.get(interface)

        self.assertIs(retrieved_entry, entry)

    def test_get_unregistered_service_raises_error(self):
        interface = Mock()

        with self.assertRaises(ServiceResolutionError) as cm:
            self.collection.get(interface)

        self.assertEqual(
            str(cm.exception),
            f"No service registered for: {interface}",
        )

    def test_set_duplicate_service_raises_error(self):
        interface = Mock()
        entry = ServiceEntry(constructor=Mock(), lifetime=Mock())

        self.collection.set(interface, entry)

        with self.assertRaises(ServiceRegistrationError) as cm:
            self.collection.set(interface, entry)

        self.assertEqual(
            str(cm.exception),
            f"Service '{interface}' is already registered.",
        )

    def test_remove_service(self):
        interface = Mock()
        entry = ServiceEntry(constructor=Mock(), lifetime=Mock())

        self.collection.set(interface, entry)
        self.collection.remove(interface)

        with self.assertRaises(ServiceResolutionError):
            self.collection.get(interface)

    def test_contains_service(self):
        interface = Mock()
        entry = ServiceEntry(constructor=Mock(), lifetime=Mock())

        self.collection.set(interface, entry)
        self.assertIn(interface, self.collection)

        self.collection.remove(interface)
        self.assertNotIn(interface, self.collection)

    def test_thread_safety(self):
        def worker(collection, interface, entry):
            collection.set(interface, entry)
            self.assertIn(interface, collection)
            retrieved_entry = collection.get(interface)
            self.assertIs(retrieved_entry, entry)

        threads = []
        for i in range(10):
            interface = type(f"MockInterface{i}", (), {})
            entry = ServiceEntry(constructor=Mock(), lifetime=Mock())
            thread = threading.Thread(target=worker, args=(self.collection, interface, entry))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        self.assertEqual(len(self.collection._entries), 10)  # Validate internal state



import unittest
from unittest.mock import Mock

from pyjudo.core import IServiceScope, IScopeStack, IResolver, IServiceCache
from pyjudo.disposable import Disposable
from pyjudo.service_scope import ServiceScope


class TestServiceScope(unittest.TestCase):
    def setUp(self):
        self.mock_cache = Mock(spec=IServiceCache)
        self.mock_stack = Mock(spec=IScopeStack)
        self.mock_resolver = Mock(spec=IResolver)
        self.scope = ServiceScope(self.mock_cache, self.mock_stack, self.mock_resolver)

    def test_get_resolves_instance(self):
        interface = Mock()
        instance = Mock()
        self.mock_resolver.resolve.return_value = instance

        result = self.scope.get(interface, key="value")
        self.assertEqual(result, instance)
        self.mock_resolver.resolve.assert_called_once_with(interface, {"key": "value"})

    def test_get_instance_retrieves_from_cache(self):
        interface = Mock()
        instance = Mock()
        self.mock_cache.get.return_value = instance

        result = self.scope.get_instance(interface)
        self.assertEqual(result, instance)
        self.mock_cache.get.assert_called_once_with(interface)

    def test_add_instance_stores_in_cache_and_registers_disposable(self):
        interface = Mock()
        instance = Mock(spec=Disposable)
        self.scope.add_instance(interface, instance)

        self.mock_cache.add.assert_called_once_with(interface, instance)
        self.assertIn(instance, self.scope.disposables)

    def test_add_instance_non_disposable(self):
        interface = Mock()
        instance = Mock()

        self.scope.add_instance(interface, instance)

        self.mock_cache.add.assert_called_once_with(interface, instance)
        self.assertNotIn(instance, self.scope.disposables)

    def test_item_access_uses_get(self):
        interface = Mock()
        self.scope.get = Mock(return_value="mock_instance")
        getter = self.scope[interface]

        result = getter(key="value")
        self.assertEqual(result, "mock_instance")
        self.scope.get.assert_called_once_with(interface, key="value")

    def test_enter_pushes_to_stack(self):
        with self.scope as entered_scope:
            self.assertEqual(entered_scope, self.scope)
            self.mock_stack.push.assert_called_once_with(self.scope)

    def test_exit_pops_stack_and_disposes_disposables(self):
        disposable1 = Mock(spec=Disposable)
        disposable2 = Mock(spec=Disposable)
        self.scope.disposables.extend([disposable1, disposable2])

        with self.scope:
            pass  # Exiting the scope triggers __exit__

        self.mock_stack.pop.assert_called_once()
        disposable1.dispose.assert_called_once()
        disposable2.dispose.assert_called_once()



"""
This type stub file was generated by pyright.
"""

from typing import Any, Protocol, override
from pyjudo.core import IResolver

class Factory[T](Protocol):
    def __call__(self, **overrides: Any) -> T:
        ...
    


class FactoryProxy[T](Factory[T]):
    """
    A proxy for a factory that resolves an interface.
    Calling the proxy will resolve the interface from the container.
    """
    _resolver: IResolver
    _interface: type[T]
    def __init__(self, resolver: IResolver, interface: type[T]) -> None:
        ...
    
    @override
    def __call__(self, **overrides: Any) -> T:
        ...
    
    @override
    def __repr__(self) -> str:
        ...
    





from abc import ABC, abstractmethod
from typing import Any, Callable
import unittest

from pyjudo.core import IResolver
from pyjudo.factory import FactoryProxy


class IFooService(ABC): ...


class FooService(IFooService): ...


class MockResolver(IResolver):
    services: dict[Any, Any] = {IFooService: FooService}

    def resolve[T](
        self, interface: type[T], overrides: dict[str, Any], binding: Any | None = None
    ) -> T:
        overrides = overrides or {}
        return self.services[interface](**overrides)

    def resolve_anonymous[T](
        self,
        constructor: Callable[..., T],
        overrides: dict[str, Any],
        binding: Any | None = None,
    ) -> T:
        raise NotImplementedError()


class Test_Factory(unittest.TestCase):
    def setUp(self):
        self.factory = FactoryProxy(MockResolver(), IFooService) # pyright: ignore[reportUninitializedInstanceVariable]

    def test_call_factory(self):
        foo = self.factory()

        self.assertIsInstance(foo, FooService)



from typing import override
import gc
import unittest
from unittest.mock import MagicMock, patch

from pyjudo import disposable
from pyjudo.disposable import IDisposable
from pyjudo.exceptions import ServiceDisposedError


class DisposableClass(IDisposable):
    def __init__(self):
        self.x: int = 1
        self.disposed_values: list[int] = []

    def get_x(self) -> int:
        return self.x

    @override
    def do_dispose(self) -> None:
        self.disposed_values.append(self.x)
        self.x = None # pyright: ignore[reportAttributeAccessIssue]


class Test_Disposable(unittest.TestCase):
    @override
    def setUp(self) -> None:
        self.disposable: IDisposable = DisposableClass() #pyright: ignore[reportUninitializedInstanceVariable]

    def test_initial_state_not_disposed(self):
        """Test that the object is not disposed upon initialization."""
        self.assertFalse(self.disposable.is_disposed)

    def test_dispose_sets_is_disposed(self):
        """Test that calling dispose sets is_disposed to True."""
        self.disposable.dispose()
        self.assertTrue(self.disposable.is_disposed)

    def test_do_dispose_called_once(self):
        """Test that do_dispose is called exactly once upon disposal."""
        with patch.object(DisposableClass, 'do_dispose') as mock_do_dispose:
            disposable = DisposableClass()
            disposable.dispose()
            mock_do_dispose.assert_called_once()

            # Call dispose again
            disposable.dispose()
            mock_do_dispose.assert_called_once()  # Still only one call

    def test_attribute_access_raises_after_disposed(self):
        """Test that accessing disposed object raises an exception."""
        self.disposable.dispose()
        with self.assertRaises(ServiceDisposedError):
            _ = self.disposable.x

    def test_attribute_set_raises_after_disposed(self):
        """Test that setting an attribute on a disposed object raises an exception."""
        self.disposable.dispose()
        with self.assertRaises(ServiceDisposedError):
            self.disposable.x = 2
    
    def test_method_call_raises_after_disposed(self):
        """Test that calling a method on a disposed object raises an exception."""
        self.disposable.dispose()
        with self.assertRaises(ServiceDisposedError):
            _ = self.disposable.get_x()

    def test_dispose_multiple_times(self):
        """Test that multiple dispose calls do not raise errors and only dispose once."""
        self.disposable.dispose()
        self.disposable.dispose()
        self.disposable.dispose()
        self.assertTrue(self.disposable.is_disposed)

    def test_dispose_cleanup(self):
        """Test that dispose cleans up resources as expected."""
        self.assertEqual(self.disposable.x, 1)
        ref = self.disposable.__dict__
        self.disposable.dispose()
        self.assertIsNone(ref["x"])
        self.assertIn(1, ref["disposed_values"])

    def test_del_calls_dispose(self):
        """Test that __del__ calls dispose if not already disposed."""
        
        with patch.object(DisposableClass, 'dispose') as mock_dispose:
            disposable = DisposableClass()
            del disposable
            _ = gc.collect()  # Force garbage collection
            mock_dispose.assert_called_once()


from functools import wraps
import inspect
import logging
import threading
from typing import TypeAlias, override, Any, Callable, Self, Union

from pyjudo.core import (
    IServiceContainer,
    IServiceEntryCollection,
    IServiceCache,
    IServiceScope,
    IScopeStack,
    IResolver,
    ServiceLife,
    ServiceEntry,
)
from pyjudo.exceptions import ServiceRegistrationError, ServiceResolutionError, ServiceTypeError
from pyjudo.factory import FactoryProxy


class InjectDecorator:
    def __init__(self, resolver: IResolver, func: Callable[..., Any]) -> None:
        self.resolver: IResolver = resolver
        self.func: Callable[..., Any] = func
    
    
    def __get__[T](self, instance: T | None, owner: type[T] | None) -> Callable[..., Any]:
        if isinstance(self.func, classmethod):
            original_func = self.func.__func__

            @wraps(original_func)
            def classmethod_wrapper(*args: Any, **kwargs: Any) -> Any:
                return self.resolver.resolve_anonymous(original_func, kwargs, binding=owner)
            return classmethod_wrapper

        elif isinstance(self.func, staticmethod):
            original_func = self.func.__func__

            @wraps(original_func)
            def staticmethod_wrapper(*args: Any, **kwargs: Any) -> Any:
                return self.resolver.resolve_anonymous(original_func, kwargs, binding=None)
            return staticmethod_wrapper

        else:
            @wraps(self.func)
            def bound_wrapper(*args: Any, **kwargs: Any) -> Any:
                return self.resolver.resolve_anonymous(self.func, kwargs, binding=instance)
            return bound_wrapper

    @property
    def __name__(self) -> str:
        return self.func.__name__

    def __call__(self, *args, **kwargs):
        return self.resolver.resolve_anonymous(self.func, kwargs)


class ServiceContainer(IServiceContainer):
    def __init__(
        self,
        service_entry_collection: IServiceEntryCollection,
        singleton_cache: IServiceCache,
        scope_stack: IScopeStack,
        scope_factory: Callable[[IScopeStack, IResolver], IServiceScope],
        resolver: IResolver,
    ) -> None:
        self.__lock = threading.Lock()
        self._logger = logging.getLogger(self.__class__.__name__)

        self.service_entry_collection = service_entry_collection
        self.singleton_cache = singleton_cache
        self.scope_stack = scope_stack
        self.scope_factory = scope_factory
        self.resolver = resolver

        self.register(IServiceContainer, self.__class__, ServiceLife.SINGLETON)
        self.singleton_cache.add(IServiceContainer, self)

    @override
    def register[T](
        self,
        interface: type[T],
        constructor: type[T] | Callable[..., T],
        lifetime: ServiceLife = ServiceLife.TRANSIENT,
    ) -> None:
        if inspect.isclass(constructor):
            if not issubclass(constructor, interface):
                raise ServiceRegistrationError(f"'{constructor}' does not implement '{interface}'")
        elif callable(constructor):
            return_annotation = inspect.signature(constructor).return_annotation

            if return_annotation is inspect.Signature.empty:
                raise ServiceRegistrationError(f"Callable '{constructor}' must have a return annotation.")
            
            if not issubclass(return_annotation, interface):
                raise ServiceRegistrationError(f"'{constructor}' does not return '{interface}'")
        else:
            raise ServiceRegistrationError("Constructor must be a class or callable")

        with self.__lock:
            entry = ServiceEntry(constructor, lifetime)
            self.service_entry_collection.set(interface, entry)

    def inject(self, func: Callable[..., Any]) -> Callable[..., Any]:
        return InjectDecorator(self.resolver, func)

    @override
    def unregister(self, interface: type) -> None:
        self.service_entry_collection.remove(interface)

    @override
    def add_transient[T](
        self, interface: type[T], constructor: type[T] | Callable[..., T]
    ) -> Self:
        self.register(interface, constructor, ServiceLife.TRANSIENT)
        return self

    @override
    def add_scoped[T](
        self, interface: type[T], constructor: type[T] | Callable[..., T]
    ) -> Self:
        self.register(interface, constructor, ServiceLife.SCOPED)
        return self

    @override
    def add_singleton[T](
        self, interface: type[T], constructor: type[T] | Callable[..., T]
    ) -> Self:
        self.register(interface, constructor, ServiceLife.SINGLETON)
        return self

    @override
    def get[T](self, interface: type[T], **overrides: Any) -> T:
        with self.__lock:
            service = self.resolver.resolve(interface, overrides)
        if not issubclass(type(service), interface):
            raise ServiceTypeError(f"Service '{service}' is not of type '{interface}'")
        return service

    @override
    def get_factory[T](self, interface: type[T]) -> Callable[..., T]:
        if not self.is_registered(interface):
            raise ServiceResolutionError(f"Service '{interface}' is not registered.")
        return FactoryProxy(self.resolver, interface)

    @override
    def is_registered(self, interface: type) -> bool:
        return interface in self.service_entry_collection

    @override
    def create_scope(self) -> IServiceScope:
        return self.scope_factory(self.scope_stack, self.resolver)



"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod

class IServiceCache(ABC):
    @abstractmethod
    def get[T](self, interface: type[T]) -> T | None:
        ...
    
    @abstractmethod
    def add[T](self, interface: type[T], instance: T) -> None:
        ...
    





import threading
from typing import Any, override

from pyjudo.core import IServiceEntryCollection, ServiceEntry
from pyjudo.exceptions import ServiceRegistrationError, ServiceResolutionError


class ServiceEntryCollection(IServiceEntryCollection):
    def __init__(self):
        self.__lock = threading.Lock()
        self._entries: dict[type[Any], ServiceEntry[Any]] = {}

    @override
    def get[T](self, interface: type[T]) -> ServiceEntry[T]:
        with self.__lock:
            entry = self._entries.get(interface)
        if entry is None:
            raise ServiceResolutionError(
                f"No service registered for: {interface}"
            )
        return entry

    @override
    def set[T](self, interface: type[T], entry: ServiceEntry[T]) -> None:
        with self.__lock:
            if interface in self._entries:
                raise ServiceRegistrationError(
                    f"Service '{interface}' is already registered."
                )
            self._entries[interface] = entry

    @override
    def remove[T](self, key: type[T]) -> None:
        _ = self._entries.pop(key, None)

    @override
    def __contains__(self, key: type) -> bool:
        return key in self._entries


"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod
from typing import Any, Callable, Self
from .service_life import ServiceLife
from .iservice_scope import IServiceScope

class IServiceContainer(ABC):
    @abstractmethod
    def register[T](self, interface: type[T], constructor: type[T] | Callable[..., T], lifetime: ServiceLife = ...) -> None:
        ...
    
    @abstractmethod
    def unregister(self, interface: type) -> None:
        ...
    
    @abstractmethod
    def add_transient[T](self, interface: type[T], constructor: type[T] | Callable[..., T]) -> Self:
        ...
    
    @abstractmethod
    def add_scoped[T](self, interface: type[T], constructor: type[T] | Callable[..., T]) -> Self:
        ...
    
    @abstractmethod
    def add_singleton[T](self, interface: type[T], constructor: type[T] | Callable[..., T]) -> Self:
        ...
    
    @abstractmethod
    def get[T](self, interface: type[T], **overrides: Any) -> T:
        ...
    
    @abstractmethod
    def get_factory[T](self, interface: type[T]) -> Callable[..., T]:
        ...
    
    @abstractmethod
    def is_registered(self, interface: type) -> bool:
        ...
    
    @abstractmethod
    def create_scope(self) -> IServiceScope:
        ...
    
    def __getitem__[T](self, key: type[T]) -> Callable[..., T]:
        ...
    





"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod
from typing import Any, Protocol, override, runtime_checkable

@runtime_checkable
class Disposable(Protocol):
    def dispose(self) -> None:
        ...
    


class IDisposable(ABC):
    """
    Represents an object that can be disposed.
    """
    __disposed: bool = ...
    @abstractmethod
    def do_dispose(self) -> None:
        """
        Custom dispose logic should be implemented in subclasses.
        """
        ...
    
    def dispose(self) -> None:
        """
        Disposes of the object, setting its disposed state to True and calling the custom dispose logic.
        """
        ...
    
    @property
    def is_disposed(self) -> bool:
        """
        Returns whether the object has been disposed.
        """
        ...
    
    def __getattr__(self, name: str): # -> Any:
        """
        Checks if the object is disposed before accessing any attributes.
        """
        ...
    
    @override
    def __getattribute__(self, name: str): # -> Any:
        """
        Checks if the object is disposed before accessing any attributes.
        """
        ...
    
    @override
    def __setattr__(self, name: str, value: Any) -> None:
        """
        Checks if the object is disposed before setting any attributes.
        """
        ...
    
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """
        Checks if the object is disposed before calling it.
        """
        ...
    
    @override
    def __str__(self) -> str:
        """
        Checks if the object is disposed before converting it to a string.
        """
        ...
    
    @override
    def __repr__(self) -> str:
        """
        Checks if the object is disposed before getting its representation.
        """
        ...
    
    def __del__(self) -> None:
        """
        Ensures dispose is called before garbage collection if not already done.
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

from collections.abc import Callable
from .service_life import ServiceLife

class ServiceEntry[T]:
    """
    Represents a service entry in the container.
    """
    __slots__ = ...
    constructor: type[T] | Callable[..., T]
    lifetime: ServiceLife
    def __init__(self, constructor: type[T] | Callable[..., T], lifetime: ServiceLife) -> None:
        ...
    





from functools import partial
import logging
from typing import override, Any, Callable

from pyjudo.core import IServiceScope, IScopeStack, IResolver, IServiceCache
from pyjudo.disposable import Disposable


class ServiceScope(IServiceScope):
    """
    Represents a scope for services.
    """

    def __init__(self, cache: IServiceCache, stack: IScopeStack, resolver: IResolver) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

        self.cache: IServiceCache = cache
        self.stack: IScopeStack = stack
        self.resolver: IResolver = resolver
        self.disposables: list[Disposable] = []

    @override
    def get[T](self, interface: type[T], **overrides: Any) -> T:
        return self.resolver.resolve(interface, overrides)

    @override
    def get_instance[T](self, interface: type[T]) -> T | None:
        return self.cache.get(interface)

    @override
    def add_instance[T](self, interface: type[T], instance: T) -> None:
        self.cache.add(interface, instance)
        if isinstance(instance, Disposable):
            self.disposables.append(instance)

    @override
    def __getitem__[T](self, key: type[T]) -> Callable[..., T]:
        return partial(self.get, key)

    @override
    def __enter__(self):
        self.stack.push(self)  # Use container's scope_stack
        return self

    @override
    def __exit__(self, _exc_type, _exc_value, _traceback):
        for disposable in self.disposables:
            disposable.dispose()
        self.stack.pop()


"""
This type stub file was generated by pyright.
"""

from typing import override
from pyjudo.core import IServiceEntryCollection, ServiceEntry

class ServiceEntryCollection(IServiceEntryCollection):
    def __init__(self) -> None:
        ...
    
    @override
    def get[T](self, interface: type[T]) -> ServiceEntry[T]:
        ...
    
    @override
    def set[T](self, interface: type[T], entry: ServiceEntry[T]) -> None:
        ...
    
    @override
    def remove[T](self, key: type[T]) -> None:
        ...
    
    @override
    def __contains__(self, key: type) -> bool:
        ...
    





"""
This type stub file was generated by pyright.
"""

from typing import Any, Callable, override
from pyjudo.core import IResolver, IScopeStack, IServiceCache, IServiceEntryCollection

class Resolver(IResolver):
    """
    Resolves services from the container.
    The resolver is responsible for retrieving service instances from a container.
    """
    def __init__(self, service_entry_collection: IServiceEntryCollection, singleton_cache: IServiceCache, scope_stack: IScopeStack) -> None:
        ...
    
    @override
    def resolve_anonymous[T](self, constructor: Callable[..., T], overrides: dict[str, Any], binding: Any | None = ...) -> T:
        ...
    
    @override
    def resolve[T](self, interface: type[T], overrides: dict[str, Any]) -> T:
        ...
    





"""
This type stub file was generated by pyright.
"""

from .iresolver import IResolver
from .iservice_cache import IServiceCache
from .iservice_container import IServiceContainer
from .iservice_entry_collection import IServiceEntryCollection
from .iservice_scope import IServiceScope
from .iservice_stack import IScopeStack
from .service_life import ServiceLife
from .service_entry import ServiceEntry

__all__ = ["IResolver", "IServiceCache", "IServiceContainer", "IServiceEntryCollection", "IServiceScope", "IScopeStack", "ServiceLife", "ServiceEntry"]



"""
This type stub file was generated by pyright.
"""

from typing import Any, Callable, Self, override
from pyjudo.core import IResolver, IScopeStack, IServiceCache, IServiceContainer, IServiceEntryCollection, IServiceScope, ServiceLife

class InjectDecorator:
    def __init__(self, resolver: IResolver, func: Callable[..., Any]) -> None:
        ...
    
    def __get__[T](self, instance: T | None, owner: type[T] | None) -> Callable[..., Any]:
        ...
    
    @property
    def __name__(self) -> str:
        ...
    
    def __call__(self, *args, **kwargs): # -> Any:
        ...
    


class ServiceContainer(IServiceContainer):
    def __init__(self, service_entry_collection: IServiceEntryCollection, singleton_cache: IServiceCache, scope_stack: IScopeStack, scope_factory: Callable[[IScopeStack, IResolver], IServiceScope], resolver: IResolver) -> None:
        ...
    
    @override
    def register[T](self, interface: type[T], constructor: type[T] | Callable[..., T], lifetime: ServiceLife = ...) -> None:
        ...
    
    def inject(self, func: Callable[..., Any]) -> Callable[..., Any]:
        ...
    
    @override
    def unregister(self, interface: type) -> None:
        ...
    
    @override
    def add_transient[T](self, interface: type[T], constructor: type[T] | Callable[..., T]) -> Self:
        ...
    
    @override
    def add_scoped[T](self, interface: type[T], constructor: type[T] | Callable[..., T]) -> Self:
        ...
    
    @override
    def add_singleton[T](self, interface: type[T], constructor: type[T] | Callable[..., T]) -> Self:
        ...
    
    @override
    def get[T](self, interface: type[T], **overrides: Any) -> T:
        ...
    
    @override
    def get_factory[T](self, interface: type[T]) -> Callable[..., T]:
        ...
    
    @override
    def is_registered(self, interface: type) -> bool:
        ...
    
    @override
    def create_scope(self) -> IServiceScope:
        ...
    





import threading
import unittest
from unittest.mock import Mock

from pyjudo.core import IServiceScope
from pyjudo.core.iservice_stack import IScopeStack
from pyjudo.exceptions import ServiceScopeError
from pyjudo.scope_stack import ScopeStack


class Test_ScopeStack(unittest.TestCase):
    def setUp(self):
        self.scope_stack: IScopeStack = ScopeStack() # pyright: ignore[reportUninitializedInstanceVariable]
        self.mock_scope: IServiceScope = Mock(spec=IServiceScope) # pyright: ignore[reportUninitializedInstanceVariable]

    def test_push_scope(self):
        self.scope_stack.push(self.mock_scope)

        self.assertEqual(self.scope_stack.get_current(), self.mock_scope)

    def test_pop_scope(self):
        self.scope_stack.push(self.mock_scope)
        self.scope_stack.pop()

        self.assertIsNone(self.scope_stack.get_current())

    def test_get_current_no_scope(self):
        self.assertIsNone(self.scope_stack.get_current())

    def test_pop_empty_stack(self):
        with self.assertRaises(ServiceScopeError):
            self.scope_stack.pop()

    def test_get_current_returns_last_pushed_scope(self):
        scope1 = Mock(spec=IServiceScope)
        scope2 = Mock(spec=IServiceScope)

        self.scope_stack.push(scope1)
        self.scope_stack.push(scope2)

        self.assertEqual(self.scope_stack.get_current(), scope2)

    def test_thread_safety(self):
        def worker(scope_stack, scope):
            scope_stack.push(scope)
            self.assertEqual(scope_stack.get_current(), scope)
            scope_stack.pop()

        threads = []
        for i in range(5):
            thread_scope = Mock(spec=IServiceScope)
            thread = threading.Thread(target=worker, args=(self.scope_stack, thread_scope))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        self.assertIsNone(self.scope_stack.get_current())


from abc import ABC
import unittest
from unittest.mock import MagicMock, Mock
from typing import Any, Callable

from pyjudo.core import IScopeStack, IServiceCache, IServiceEntryCollection, IServiceScope, IResolver, ServiceLife
from pyjudo.exceptions import ServiceRegistrationError, ServiceResolutionError, ServiceTypeError
from pyjudo.factory import FactoryProxy
from pyjudo.service_container import ServiceContainer

class IMockService(ABC): ...

class MockService(IMockService): ...

class Test_ServiceContainerInject(unittest.TestCase):
    def setUp(self):
        self.mock_service_entry_collection: IServiceEntryCollection = MagicMock(spec=IServiceEntryCollection)
        self.mock_singleton_cache: IServiceCache = Mock(spec=IServiceCache)
        self.mock_scope_stack: IScopeStack = Mock(spec=IScopeStack)
        self.mock_resolver: IResolver = Mock(spec=IResolver)
        self.mock_scope_factory: Callable[..., IServiceScope] = Mock(return_value=Mock(spec=IServiceScope))

        self.container = ServiceContainer(
            service_entry_collection=self.mock_service_entry_collection,
            singleton_cache=self.mock_singleton_cache,
            scope_stack=self.mock_scope_stack,
            scope_factory=self.mock_scope_factory,
            resolver=self.mock_resolver,
        )

    def test_inject_decorator(self):

        mock_service = MockService()
        self.mock_resolver.resolve_anonymous.return_value = mock_service

        self.container.register(IMockService, MockService, ServiceLife.SINGLETON)
        
        @self.container.inject
        def test_function(service: IMockService):
            return service

        rtn = test_function()
        self.assertEqual(rtn, mock_service)

    def test_inject_decorator_instance_method(self):
        mock_service = MockService()
        self.mock_resolver.resolve_anonymous.return_value = mock_service

        self.container.register(IMockService, MockService, ServiceLife.SINGLETON)

        class SomeClass:
            @self.container.inject
            def method(self, service: IMockService):
                return service

        instance = SomeClass()
        rtn = instance.method()
        self.assertEqual(rtn, mock_service)

    def test_inject_decorator_classmethod(self):
        mock_service = MockService()
        self.mock_resolver.resolve_anonymous.return_value = mock_service

        self.container.register(IMockService, MockService, ServiceLife.SINGLETON)

        class SomeClass:
            @self.container.inject
            @classmethod
            def method(cls, service: IMockService):
                return service

        rtn = SomeClass.method()
        self.assertEqual(rtn, mock_service)

    def test_inject_decorator_staticmethod(self):
        mock_service = MockService()
        self.mock_resolver.resolve_anonymous.return_value = mock_service

        self.container.register(IMockService, MockService, ServiceLife.SINGLETON)

        class SomeClass:
            @self.container.inject
            @staticmethod
            def method(service: IMockService):
                return service

        rtn = SomeClass.method()
        self.assertEqual(rtn, mock_service)


import threading
from typing import override, Any


from pyjudo.core import IServiceCache


class ServiceCache(IServiceCache):
    def __init__(self, initial: dict[type[Any], Any] | None = None):
        self.__lock = threading.Lock()
        self._cache: dict[type[Any], Any] = initial or {}

    @override
    def get[T](self, interface: type[T]) -> T | None:
        with self.__lock:
            return self._cache.get(interface)

    @override
    def add[T](self, interface: type[T], instance: T) -> None:
        with self.__lock:
            self._cache[interface] = instance


"""
This type stub file was generated by pyright.
"""

class ServiceException(Exception):
    ...


class ServiceCircularDependencyError(ServiceException):
    ...


class ServiceResolutionError(ServiceException):
    ...


class ServiceRegistrationError(ServiceException):
    ...


class ServiceTypeError(ServiceException):
    ...


class ServiceDisposedError(ServiceException):
    ...


class ServiceScopeError(ServiceException):
    ...





"""
This type stub file was generated by pyright.
"""

from typing import override
from pyjudo.core import IScopeStack, IServiceScope

class ScopeStack(IScopeStack):
    def __init__(self) -> None:
        ...
    
    @override
    def get_current(self) -> IServiceScope | None:
        ...
    
    @override
    def push(self, scope: IServiceScope) -> None:
        ...
    
    @override
    def pop(self) -> None:
        ...
    





import inspect
from typing import Any, Callable, Protocol, runtime_checkable
import unittest
from unittest.mock import MagicMock, Mock, patch

from pyjudo.core import (
    IServiceEntryCollection,
    IServiceCache,
    IServiceScope,
    IScopeStack,
    ServiceLife,
)
from pyjudo.exceptions import ServiceCircularDependencyError, ServiceResolutionError, ServiceScopeError
from pyjudo.factory import Factory, FactoryProxy
from pyjudo.resolver import Resolver

@runtime_checkable
class IServiceA(Protocol):
    pass


class ServiceA(IServiceA):
    pass


@runtime_checkable
class IServiceB(Protocol):
    pass


class ServiceB(IServiceB):
    def __init__(self, service_a: IServiceA):
        self.service_a = service_a

class Test_Resolver(unittest.TestCase):
    def setUp(self):
        self.mock_service_entry_collection: IServiceEntryCollection = MagicMock(spec=IServiceEntryCollection) # pyright: ignore[reportUninitializedInstanceVariable]
        self.mock_service_cache: IServiceCache = Mock(spec=IServiceCache) # pyright: ignore[reportUninitializedInstanceVariable]
        self.mock_service_scope_stack: IScopeStack = Mock(spec=IScopeStack) # pyright: ignore[reportUninitializedInstanceVariable]

        self.resolver: Resolver = Resolver( # pyright: ignore[reportUninitializedInstanceVariable]
            self.mock_service_entry_collection,
            self.mock_service_cache,
            self.mock_service_scope_stack,
        )

    def test_resolution_stack_initialisation(self):
        stack = self.resolver._resolution_stack
        self.assertEqual(stack, set())

    def test_resolve_anonymous(self):
        constructor = Mock(return_value="instance")
        instance = self.resolver.resolve_anonymous(constructor, {})

        self.assertEqual(instance, "instance")

    def test_resolve_singleton(self):
        interface = Mock()
        entry = Mock()
        entry.lifetime = ServiceLife.SINGLETON
        entry.constructor = Mock(return_value="singleton_instance")
        self.mock_service_entry_collection.get.return_value = entry

        # Simulate instance not cached
        self.mock_service_cache.get.return_value = None

        instance = self.resolver.resolve(interface, {})

        self.assertEqual(instance, "singleton_instance")
        self.mock_service_cache.add.assert_called_once()

    def test_resolve_singleton_cached(self):
        interface = Mock()
        entry = Mock()
        entry.lifetime = ServiceLife.SINGLETON
        entry.constructor = Mock(return_value="singleton_instance")
        self.mock_service_entry_collection.get.return_value = entry

        # Simulate instance cached
        self.mock_service_cache.get.return_value = "cached_instance"

        instance = self.resolver.resolve(interface, {})

        self.assertEqual(instance, "cached_instance")
        self.mock_service_cache.add.assert_not_called()

    def test_resolve_scoped(self):
        interface = Mock()
        entry = Mock()
        entry.lifetime = ServiceLife.SCOPED
        entry.constructor = Mock(return_value="scoped_instance")
        self.mock_service_entry_collection.get.return_value = entry
        
        # Simulate instance not cached
        current_scope = Mock(spec=IServiceScope)
        current_scope.get_instance.return_value = None

        self.mock_service_scope_stack.get_current.return_value = current_scope

        instance = self.resolver.resolve(interface, {})

        self.assertEqual(instance, "scoped_instance")
        current_scope.add_instance.assert_called_once()

    def test_resolve_scoped_cached(self):
        interface = Mock()
        entry = Mock()
        entry.lifetime = ServiceLife.SCOPED
        entry.constructor = Mock(return_value="scoped_instance")
        self.mock_service_entry_collection.get.return_value = entry

        current_scope = Mock(spec=IServiceScope)
        current_scope.get_instance.return_value = "cached_instance"

        self.mock_service_scope_stack.get_current.return_value = current_scope

        instance = self.resolver.resolve(interface, {})

        self.assertEqual(instance, "cached_instance")
        current_scope.add_instance.assert_not_called()

    def test_resolve_transient(self):
        interface = Mock()
        entry = Mock()
        entry.lifetime = ServiceLife.TRANSIENT
        entry.constructor = Mock(return_value="transient_instance")
        self.mock_service_entry_collection.get.return_value = entry

        instance = self.resolver.resolve(interface, {})

        self.assertEqual(instance, "transient_instance")

    def test_circular_dependency(self):
        interface = Mock()

        self.resolver._resolution_stack.add(interface)

        with self.assertRaises(ServiceCircularDependencyError):
            self.resolver.resolve(interface, {})

    def test_no_scope_error(self):
        interface = Mock()
        entry = Mock()
        entry.lifetime = ServiceLife.SCOPED
        self.mock_service_entry_collection.get.return_value = entry

        self.mock_service_scope_stack.get_current.return_value = None

        with self.assertRaises(ServiceScopeError):
            self.resolver.resolve(interface, {})

    def test_singleton_overrides_error(self):
        interface = Mock()
        entry = Mock()
        entry.lifetime = ServiceLife.SINGLETON
        self.mock_service_entry_collection.get.return_value = entry

        self.mock_service_cache.get.return_value = "cached_instance"

        with self.assertRaises(ServiceResolutionError):
            self.resolver.resolve(interface, {"override": "value"})

    def test_create_instance_with_callable_constructor(self):
        # Test when constructor is a function
        def constructor(service_a: IServiceA):
            return ServiceB(service_a)

        self.mock_service_entry_collection.__contains__.side_effect = lambda x: x == IServiceA
        self.mock_service_entry_collection.get.return_value = Mock(
            constructor=ServiceA, lifetime=ServiceLife.TRANSIENT
        )
        service_instance = self.resolver._create_instance(constructor, overrides={})
        self.assertIsInstance(service_instance, ServiceB)
        self.assertIsInstance(service_instance.service_a, ServiceA)

    def test_create_instance_with_positional_only_parameter(self):
        # Test with positional-only parameters (Python 3.8+)
        def constructor(service_a: IServiceA, /):
            return ServiceB(service_a)

        self.mock_service_entry_collection.__contains__.side_effect = lambda x: x == IServiceA
        self.mock_service_entry_collection.get.return_value = Mock(
            constructor=ServiceA, lifetime=ServiceLife.TRANSIENT
        )
        service_instance = self.resolver._create_instance(constructor, overrides={})
        self.assertIsInstance(service_instance, ServiceB)

    def test_create_instance_with_keyword_only_parameter(self):
        # Test with keyword-only parameters
        def constructor(*, service_a: IServiceA):
            return ServiceB(service_a)

        self.mock_service_entry_collection.__contains__.side_effect = lambda x: x == IServiceA
        self.mock_service_entry_collection.get.return_value = Mock(
            constructor=ServiceA, lifetime=ServiceLife.TRANSIENT
        )
        service_instance = self.resolver._create_instance(constructor, overrides={})
        self.assertIsInstance(service_instance, ServiceB)

    def test_create_instance_with_var_positional_parameters(self):
        # Test that *args are skipped
        def constructor(*args, service_a: IServiceA):
            return ServiceB(service_a)

        self.mock_service_entry_collection.__contains__.side_effect = lambda x: x == IServiceA
        self.mock_service_entry_collection.get.return_value = Mock(
            constructor=ServiceA, lifetime=ServiceLife.TRANSIENT
        )
        service_instance = self.resolver._create_instance(constructor, overrides={})
        self.assertIsInstance(service_instance, ServiceB)

    def test_create_instance_with_var_keyword_parameters(self):
        # Test that **kwargs are skipped
        def constructor(service_a: IServiceA, **kwargs):
            return ServiceB(service_a)

        self.mock_service_entry_collection.__contains__.side_effect = lambda x: x == IServiceA
        self.mock_service_entry_collection.get.return_value = Mock(
            constructor=ServiceA, lifetime=ServiceLife.TRANSIENT
        )
        service_instance = self.resolver._create_instance(constructor, overrides={})
        self.assertIsInstance(service_instance, ServiceB)

    def test_create_instance_with_factory_annotation(self):
        # Test parameters annotated with Factory[Interface]
        class ServiceC:
            def __init__(self, factory: Factory[IServiceA]):
                self.factory = factory

        service_instance = self.resolver._create_instance(ServiceC, overrides={})
        self.assertIsInstance(service_instance, ServiceC)
        self.assertIsInstance(service_instance.factory, FactoryProxy)

    def test_create_instance_with_overrides(self):
        # Test when parameter name is in overrides
        def constructor(service_a: IServiceA):
            return ServiceB(service_a)

        override_service = ServiceA()
        service_instance = self.resolver._create_instance(constructor, overrides={'service_a': override_service})
        self.assertIsInstance(service_instance, ServiceB)
        self.assertEqual(service_instance.service_a, override_service)

    def test_create_instance_with_default_parameter(self):
        # Test parameter has default value
        def constructor(service_a: IServiceA = ServiceA()):
            return ServiceB(service_a)

        service_instance = self.resolver._create_instance(constructor, overrides={})
        self.assertIsInstance(service_instance, ServiceB)
        self.assertIsInstance(service_instance.service_a, ServiceA)

    def test_create_instance_unresolvable_parameter_raises_error(self):
        # Test when parameter cannot be resolved
        def constructor(service_a: IServiceA):
            return ServiceB(service_a)

        self.mock_service_entry_collection.__contains__.return_value = False
        with self.assertRaises(ServiceResolutionError):
            self.resolver._create_instance(constructor, overrides={})

    def test_create_instance_with_binding_self(self):
        # Test handling 'self' when binding is provided
        class ServiceD:
            def method(self, service_a: IServiceA):
                return service_a

        instance = ServiceD()
        self.mock_service_entry_collection.__contains__.side_effect = lambda x: x == IServiceA
        self.mock_service_entry_collection.get.return_value = Mock(
            constructor=ServiceA, lifetime=ServiceLife.TRANSIENT
        )

        bound_method = instance.method
        service_instance = self.resolver._create_instance(bound_method, overrides={}, binding=instance)
        self.assertIsInstance(service_instance, ServiceA)

    def test_create_instance_constructor_is_class(self):
        # Test when constructor is a class
        self.mock_service_entry_collection.__contains__.side_effect = lambda x: False
        service_instance = self.resolver._create_instance(ServiceA, overrides={})
        self.assertIsInstance(service_instance, ServiceA)

    def test_create_instance_constructor_with_cls_binding(self):
        # Test handling 'cls' when binding is provided
        class ServiceE:
            @classmethod
            def create(cls, service_a: IServiceA):
                return cls()

        self.mock_service_entry_collection.__contains__.side_effect = lambda x: x == IServiceA
        self.mock_service_entry_collection.get.return_value = Mock(
            constructor=ServiceA, lifetime=ServiceLife.TRANSIENT
        )

        service_instance = self.resolver._create_instance(ServiceE.create, overrides={}, binding=ServiceE)
        self.assertIsInstance(service_instance, ServiceE)